import pytest
import numpy as np
import itertools

import os
import pandas as pd
import gdown
import zipfile
import moval

"""Test the func with classification tasks.

    Examples:
        >>> pytest tests/test_demo_cls.py

"""

# download the data, which we used for MICCAI 2022

output = "data_moval.zip"
if not os.path.exists(output):
    url = "https://drive.google.com/u/0/uc?id=139pqxkG2ccIFq6qNArnFJWQ2by2Spbxt&export=download"
    output = "data_moval.zip"
    gdown.download(url, output, quiet=False)

directory_data = "data_moval"
if not os.path.exists(directory_data):
    with zipfile.ZipFile(output, 'r') as zip_ref:
        zip_ref.extractall(directory_data)

# now I am playing with cifar10 classification
val_data =  "data_moval/cifar10results/predictions_val.csv"
test_data = "data_moval/cifar10results/predictions_val_motion_blur.csv"
# validation data
cnn_pred = pd.read_csv(val_data)
targets_all = np.array(cnn_pred[['target_0', 'target_1', 'target_2', 'target_3', 'target_4', 
                                 'target_5', 'target_6', 'target_7', 'target_8', 'target_9']])
logits = np.array(cnn_pred[['logit_0', 'logit_1', 'logit_2', 'logit_3', 'logit_4', 
                               'logit_5', 'logit_6', 'logit_7', 'logit_8', 'logit_9']])
gt = np.argmax(targets_all, axis = 1)
# test data
cnn_pred_test = pd.read_csv(test_data)
targets_all_test = np.array(cnn_pred_test[['target_0', 'target_1', 'target_2', 'target_3', 'target_4', 
                                           'target_5', 'target_6', 'target_7', 'target_8', 'target_9']])
logits_test = np.array(cnn_pred_test[['logit_0', 'logit_1', 'logit_2', 'logit_3', 'logit_4', 
                                      'logit_5', 'logit_6', 'logit_7', 'logit_8', 'logit_9']])
gt_test = np.argmax(targets_all_test, axis = 1)

# logits is of shape ``(n, d)``
# gt is of shape ``(n, )``

results_files = "results_demo_cls.txt"
# clean previous results
if os.path.isfile(results_files):
    os.remove(results_files)

@pytest.mark.parametrize(
        "estim_algorithm, mode, confidence_scores, class_specific", 
        list(itertools.product(moval.models.get_estim_options(),
                               ["classification"],
                               moval.models.get_conf_options(),
                               [False, True])),
)
def test_cls(estim_algorithm, mode, confidence_scores, class_specific):

    moval_model = moval.MOVAL(
                mode = mode,
                confidence_scores = confidence_scores,
                estim_algorithm = estim_algorithm,
                class_specific = class_specific
                )

    #
    moval_model.fit(logits, gt)
    estim_acc = moval_model.estimate(logits)
    pred = np.argmax(logits, axis = 1)
    err_val = np.abs( np.sum(gt == pred) / len(gt) - estim_acc )

    # save the test err in the result files.

    estim_acc_test = moval_model.estimate(logits_test)
    pred_test = np.argmax(logits_test, axis = 1)
    err_test = np.abs( np.sum(gt_test == pred_test) / len(gt_test) - estim_acc_test )

    test_condition = f"estim_algorithm = {estim_algorithm}, mode = {mode}, confidence_scores = {confidence_scores}, class_specific = {class_specific}"

    with open(results_files, 'a') as f:
        f.write(test_condition)
        f.write('\n')
        f.write("validation err: ")
        f.write(str(err_val))
        f.write('\n')
        f.write("moval parameter: ")
        f.write(str(moval_model.model_.param))
        f.write('\n')
        if moval_model.model_.extend_param:
            f.write("moval extended parameter: ")
            f.write(str(moval_model.model_.param_ext))
            f.write('\n')
        f.write("test err: ")
        f.write(str(err_test))
        f.write('\n')
        f.write('\n')

    



import pytest
import numpy as np

import moval
import moval.models

"""Test the confidence calculation functions.

    Examples:
        >>> pytest tests/test_confidences.py

"""

@pytest.mark.parametrize("confidence_scores", moval.models.get_conf_options())
def test_conf(confidence_scores):
    conf = moval.models.init(confidence_scores)
    assert isinstance(conf, moval.models.Confidence)

    # classification logit input
    inp = np.random.randn(100, 10) # 100 samples, 10 classes
    score = conf(inp)
    assert isinstance(score, np.ndarray)
    assert score.shape == (100, )

    if not conf.normalization:
        assert np.max(score) <= 1
        assert np.min(score) >= 0

    inp_large_var = 10 * inp # logit with higher confidence
    score_large_var = conf(inp_large_var)
    assert np.min(score_large_var - score) > 0
    # should be more confidence!

    # segmentation logit input
    inp = []
    for _ in range(10):
        inp.append(np.random.randn(3, 100, 100, 100)) # 10 samples, 3 classes, of volume shape ``(100, 100, 100)``
    score = conf(inp)
    assert isinstance(score, list)
    assert score[0].shape == (100, 100, 100)

    if not conf.normalization:
        assert np.max(score) <= 1
        assert np.min(score) >= 0


from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union

import numpy as np

def cal_softmax(x, T = 1, e1 = 1e-6) -> np.ndarray:
    """Compute softmax values for each sets of scores in x.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
        e1: A small number to prevent unexpected results of division.
    
    Returns:
        prob: The calculated softmax probability for different classes of shape ``(n, d)``

    """
    # to avoid exponent exploding
    x_max = np.max(x, keepdims=True, axis = 1).repeat(x.shape[1], axis=1)

    prob = np.exp((x - x_max).transpose() / T) / (np.sum(np.exp((x - x_max).transpose() / T), axis=0) + e1)

    return prob.transpose()

def cal_energy(x, T = 1) -> np.ndarray:
    """Compute energy values for each sets of scores in x.

    Calculating based on the defination as described in:
    ``https://proceedings.neurips.cc/paper/2020/hash/f5496252609c43eb8a3d147ab9b9c006-Abstract.html``
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        energy: The calculated energy of shape ``(n, )``
    
    Note:
        We do not acutally utilize T to calibrate the model here, as the score is unbounded.
        Instead, we utilize the model parameter to normalize the score.

    """
    T = 1
    denominator = np.sum(np.exp(x.transpose() / T), axis=0)
    energy = - T * np.log(denominator)

    return energy

def cal_mcp(x, T = 1) -> np.ndarray:
    """Compute maximum class probability (MCP) for each sets of scores in x.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        MCP: The calculated MCP of shape ``(n, )``

    """
    p = cal_softmax(x, T = T)
    MCP = np.max(p, axis=1)

    return MCP

def cal_entropy(x, T = 1, e1 = 1e-6) -> np.ndarray:
    """Compute entropy values for each sets of scores in x.

    Note:
        This entropy calculation chooses base as the number of class to ensure normalization.
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
        e1: A small number to prevent unexpected results of division.
    
    Returns:
        entropy: The calculated entropy of shape ``(n, )``

    """
    num_class = x.shape[-1]
    p = cal_softmax(x, T = T)
    entropy = - np.sum(p * np.emath.logn(num_class, p + e1), axis=1)

    return entropy

def cal_doctor(x, T = 1) -> np.ndarray:
    """Compute engery values for each sets of scores in x.

    Calculating based on the defination as described in:
    ``https://arxiv.org/abs/2106.02395``
    
    Args:
        x: The network output (logits) of shape ``(n, d)``.
        T: A scarlar temperature to calibrate the score.
    
    Returns:
        doctor: The calculated doctor of shape ``(n, )``
    
    Note:
        We do not acutally utilize T to calibrate the model here, as the score is unbounded.
        Instead, we utilize the model parameter to normalize the score.

    """
    T = 1
    p = cal_softmax(x, T = T)
    g = np.sum(p ** 2, axis=1)
    doctor = (1 - g) / g

    return doctor

def sum_tensor(inp, axes, keepdims=False) -> np.ndarray:
    axes = np.unique(axes).astype(int)
    if keepdims:
        for ax in axes:
            inp = inp.sum(int(ax), keepdims=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp

def get_tp_fp_fn(net_output, gt, axes=None, square=False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate the true positive (tp), false positive (fp) and false negative (fn).

    Args:
        net_output: The network output (logits) of shape ``(n, d, H, W, (D))``.
        gt: The manual label of shape ``(n, H, W, (D))``.
        axes: We average the axes of features. This is useful if we want to return case-wise soft dice score.
        square: Indicate if we want to include the square term in tp etc, which could be more stable.
    
    Returns:
        tp: The calculated class-wise true positive of shape ``(n, d)`` or ``(d, )``.
        fp: The calculated class-wise false positive of shape ``(n, d)`` or ``(d, )``.
        fn: The calculated class-wise false negative of shape ``(n, d)`` or ``(d, )``.

    """
    if axes is None:
        axes = tuple(range(2, len(net_output.size())))

    shp_x = net_output.shape
    shp_y = gt.shape

    y_onehot = one_hot_embedding(gt, shp_x[1])

    tp = net_output * y_onehot
    fp = net_output * (1 - y_onehot)
    fn = (1 - net_output) * y_onehot

    if square:
        tp = tp ** 2
        fp = fp ** 2
        fn = fn ** 2

    tp = sum_tensor(tp, axes, keepdims=False)
    fp = sum_tensor(fp, axes, keepdims=False)
    fn = sum_tensor(fn, axes, keepdims=False)

    return tp, fp, fn


def SoftDiceLoss(x, y, smooth = 1e-5) -> np.ndarray:
    '''Calculation of soft dice score.

    Args:
        x: The network output (logits) of shape ``(n, d, H, W, (D))``.
        y: The manual label of shape ``(n, H, W, (D))``.
    
    Returns:
        dice_score: Average class-wise dice of shape ``(d, )``

    '''
    assert len(x.shape) == len(y.shape) + 1

    shp_x = x.shape
    square = False

    axes = [0] + list(range(2, len(shp_x)))
    tp, fp, fn = get_tp_fp_fn(x, y, axes, square)
    dc = (2 * tp + smooth) / (2 * tp + fp + fn + smooth)
    dc_process = dc

    return dc_process

def one_hot_embedding(labels: np.ndarray, num_class: int) -> np.ndarray:
    '''Embedding labels to one-hot form.

    Args:
        labels: The label map of shape ``(n, )`` or ``(n, H, W, (D))``.
        num_class: An int scalar number of classes.

    Returns:
        labels_onehot: The transformed onehot label map of shape ``(n, d)`` or ``(n, d, H, W, (D))``.
    
    '''
    y = np.eye(num_class)[labels.reshape(-1).astype(int)]  # of shape `(n * H * W (*D), d)`
    labels_onehot = y.reshape(list(labels.shape)+[num_class]) # of shape `(n, H, W, (D), d)`

    axes = [0] + [len(labels.shape)] + list(range(1, len(labels.shape)))
    labels_onehot = np.transpose(labels_onehot, axes) # of shape `(n, d, H, W, (D))`

    return labels_onehot


"""Define the MOVAL model."""
import numpy as np
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
from sklearn.base import BaseEstimator
import moval.models
import moval.solvers
import pickle

class MOVAL(BaseEstimator):
    """ MOVAL model defined as part of a ``scikit-learn``-like API.

    Attributes:
        mode (str):
            The given task to estimate model performance. |Default:| ``classification``
        confidence_scores (str):
            The method to calculate the confidence scores. We provide a list of confidence score calculation methods which 
            can be displayed by running :py:func:`moval.models.get_conf_options`. |Default:| ``raw``
        estim_algorithm (str):
            The algorithm to estimate model performance. We also provide a list of estimation algorithm which can be displayed by
            running :py:func:`moval.models.get_estim_options`. |Default:| ``ac``
        class_specific (bool):
            If ``True``, the calculation will match class-wise confidence to class-wise accuracy. |Default:| ``False``
        approximate (bool):
            If ``True``, we crop the image and label map to accelerate the optimization of segmentation. |Default:| ``False``
        approximate_boundary (int):
            The enlarged regions of the region to crop. |Default:| ``30``

    Example:
        
        >>> import moval
        >>> moval_model = moval.MOVAL(
                            mode = "classification",
                            confidence_scores = "max_class_probability-conf",
                            estim_algorithm = "ac-model",
                            class_specific = False
                            )

    """


    def __init__(
        self,
        mode: str = "classification",
        confidence_scores: str = "max_class_probability-conf",
        estim_algorithm: str = "ac-model",
        class_specific: bool = False, 
        approximate: bool = False, 
        approximate_boundary: int = 30
    ):
        self.__dict__.update(locals())

    def fit(
        self,
        logits: Union[List[Iterable], np.ndarray],
        gt: Union[List[Iterable], np.ndarray]
    ) -> "MOVAL":
        """Fit the estimator to the given dataset by minimzing the calibration error.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` for classification and a list of n ``(d, H, W, (D))`` for segmentation. 
            gt: The cooresponding annotation of shape ``(n, )`` for classification and a list of n ``(H, W, (D))`` for segmentation.
        
        Return:
            ``self``

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)

        """

        # initilization
        if isinstance(logits, list):
            self.numclass = logits[0].shape[0]
        else:
            self.numclass = logits.shape[-1]

        model = moval.models.init(
            self.estim_algorithm,
            mode = self.mode,
            num_class = self.numclass,
            confidence_scores = self.confidence_scores,
            class_specific = self.class_specific
            )
        solver = moval.solvers.init("base-solver", model = model)

        # Input validation
        if self.mode == "classification":
            if logits.shape[0] != gt.shape[0]:
                raise ValueError(
                        f"Invalid input samples: logits and GT should represent the same number of samples"
                        f"(n_samples of logits, {logits.shape[0]}, while n_samples of GT , {gt.shape[0]})."
                    )

            if len(logits.shape) != (len(gt.shape) + 1):
                raise ValueError(
                        f"Invalid input dimension: logits should have one additional dimension than GT"
                        f"(dimension of logits, {len(logits.shape)}, while dimension of GT , {len(gt.shape)})."
                    )
            
            if len(gt.shape) != 1:
                raise ValueError(
                    f"Invalid input dimension for classification: GT should have the dimension of 1"
                    f"(dimension of GT, {len(logits.shape)})."
                )

        else:
            if len(logits) != len(gt):
                raise ValueError(
                        f"Invalid input samples: logits and GT should represent the same number of samples"
                        f"(n_samples of logits, {len(logits)}, while n_samples of GT , {len(gt)})."
                    )

            if len(logits[0].shape) != (len(gt[0].shape) + 1):
                raise ValueError(
                        f"Invalid input dimension: logits should have one additional dimension than GT"
                        f"(dimension of logits, {len(logits[0].shape)}, while dimension of GT , {len(gt[0].shape)})."
                    )
            
            if (len(gt[0].shape) != 2) and (len(gt[0].shape) != 3):
                raise ValueError(
                    f"Invalid input dimension for segmentation: GT should have the dimension of 2 or 3"
                    f"(dimension of GT, {len(gt[0].shape)})."
                )

        if self.mode == "segmentation" and self.approximate == True:
            logits, gt = self.crop(logits = logits, gt = gt, approximate_boundary = self.approximate_boundary)

        solver.fit(logits, gt)

        #
        self.model_ = model
        self.solver_ = solver
        if self.mode == "classification":
            self.n_dim_ = len(logits.shape)
        else:
            self.n_dim_ = len(logits[0].shape)

        return self

    @classmethod
    def crop(self,
             logits: List[Iterable],
             gt: List[Iterable],
             approximate_boundary):
        """Crop the image and label map to accelrate the optimization process.
        
        Here we do the cropping based on the label map (gt). We first find the bounding box and enlarge the region with boundary = 30 pixels.

        Args:
            logits: The network output (logits) of a list of n ``(d, H, W, (D))``. 
            gt: The cooresponding annotation of a list of n ``(H, W, (D))``.
            approximate_boundary: The enlarged regions of the region to crop.
        
        Return:
            logits_post: The cropped network output (logits) of a list of n ``(d, H', W', (D'))``. 
            gt_post: The cooresponding cropped annotation of a list of n ``(H', W', (D'))``.

        """

        logits_post = []
        gt_post = []
        for k_case in range(len(logits)):
            logit_map = logits[k_case]
            gt_map = gt[k_case]
        
            if len(gt_map.shape) == 2:
                border_x, border_y = gt_map.shape
                ind_x, ind_y = np.where(gt_map != 0)
                #
                max_x = np.min((np.max(ind_x) + approximate_boundary, border_x))
                min_x = np.max((np.min(ind_x) - approximate_boundary, 0))
                #
                max_y = np.min((np.max(ind_y) + approximate_boundary, border_y))
                min_y = np.max((np.min(ind_y) - approximate_boundary, 0))
                #
                logits_post.append(logit_map[:, min_x:max_x, min_y:max_y])
                gt_post.append(gt_map[min_x:max_x, min_y:max_y])

            elif len(gt_map.shape) == 3:
                border_x, border_y, border_z = gt_map.shape
                ind_x, ind_y, ind_z = np.where(gt_map != 0)
                #
                max_x = np.min((np.max(ind_x) + approximate_boundary, border_x))
                min_x = np.max((np.min(ind_x) - approximate_boundary, 0))
                #
                max_y = np.min((np.max(ind_y) + approximate_boundary, border_y))
                min_y = np.max((np.min(ind_y) - approximate_boundary, 0))
                #
                max_z = np.min((np.max(ind_z) + approximate_boundary, border_z))
                min_z = np.max((np.min(ind_z) - approximate_boundary, 0))
                #
                logits_post.append(logit_map[:, min_x:max_x, min_y:max_y, min_z:max_z])
                gt_post.append(gt_map[min_x:max_x, min_y:max_y, min_z:max_z])
            else:
                raise ValueError("Not implemented!")
        
        return logits_post, gt_post
        

    
    def estimate(self,
                 logits: Union[List[Iterable], np.ndarray],
                 gt_guide: np.ndarray = None):
        """Estimate model performance using logits.

        Args:
            logits: A numpy array of size ``(n, d)`` for classification or a list of n ``(d, H, W, (D))`` for segmentation.
            gt_guide: A numpy array of size ``(n, d)`` for segmentation, indicating the existing of object d in sample n.
                If ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim: estimated accuracy (float) for classification tasks, 
                or estimated foreground dsc of shape ``(d-1, )`` for segmentation tasks.

        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> estim_acc = moval_model.estimate(logits)

        """

        # Input validation
        if self.mode == "classification":
            if self.n_dim_ != len(logits.shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits.shape)})"
                    )
        else:
            if self.n_dim_ != len(logits[0].shape):
                raise ValueError(
                        f"Inconsistent input dimension: training and test samples have different dimenstions"
                        f"(dimension of training samples, {len(self.n_dim_)}, while dimension of test samples, {len(logits[0].shape)})"
                    )

        model = self.model_

        if model.mode == "classification":
            estim_acc, estim_acc_allcls = model(logits)
            return estim_acc
        else:
            estim_acc, estim_dsc = model(logits, gt_guide = gt_guide)
            return estim_dsc[1:]
    
    def save(self, filename: str):
        """Save current parameters to disk using pickle.

        Args:
            filename: The path to the file in which to save the parameters.
        
        Return:
            The saved parameters.

        Examples:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> moval_model.save('./foo.pkl')

        """
        # Object or variable to be saved
        ckpt = {}

        ckpt["estim_algorithm"] = self.model_.estim_algorithm
        ckpt["mode"] = self.model_.mode
        ckpt["num_class"] = self.model_.num_class
        ckpt["confidence_scores"] = self.model_.confidence_scores
        ckpt["class_specific"] = self.model_.class_specific
        ckpt["param"] = self.model_.param
        ckpt["n_dim_"] = self.n_dim_
        if self.model_.extend_param:
            ckpt["param_ext"] = self.model_.param_ext

        # Save the object as a pickle file
        with open(filename, 'wb') as file:
            pickle.dump(ckpt, file)

        return ckpt
    
    @classmethod
    def load(cls, filename: str) -> "MOVAL":
        """load parameters from disk.

        Args:
            filename: The path to the file in which to save the parameters.
        
        Return:
            The loaded model.
        
        Example:

            >>> import moval
            >>> import numpy as np
            >>> logits = np.random.randn(1000, 10)
            >>> gt = np.random.randint(0, 10, (1000))
            >>> moval_model = moval.MOVAL()
            >>> moval_model.fit(logits, gt)
            >>> moval_model.save('./foo.pkl')
            >>> loaded_model = moval.MOVAL.load('./foo.pkl')
            >>> estim_acc = loaded_model.estimate(logits)

        """

        with open(filename, 'rb') as file:
            loaded_ckpt = pickle.load(file)
        
        moval_model = moval.MOVAL(
                            mode = loaded_ckpt["mode"],
                            confidence_scores = loaded_ckpt["confidence_scores"],
                            estim_algorithm = loaded_ckpt["estim_algorithm"],
                            class_specific = False
                            )

        model = moval.models.init(
            loaded_ckpt["estim_algorithm"],
            mode = loaded_ckpt["mode"],
            num_class = loaded_ckpt["num_class"],
            confidence_scores = loaded_ckpt["confidence_scores"],
            class_specific = loaded_ckpt["class_specific"]
            )

        model.param = loaded_ckpt["param"]
        if model.extend_param:
            model.param_ext = loaded_ckpt["param_ext"]
        
        model.is_fitted = True

        moval_model.model_ = model
        moval_model.n_dim_ = loaded_ckpt["n_dim_"]

        return moval_model
    
# Dynamically remove the set_fit_request method
del MOVAL.set_fit_request


import pytest
import numpy as np
import itertools

import moval
import moval.models
import moval.solvers
import moval.registry

"""Test the confidence calculation functions.

    Examples:
        >>> pytest tests/test_solver.py

"""

def test_registry():
    assert moval.registry.is_registry(moval.solvers)
    assert moval.registry.is_registry(moval.solvers, check_docs=True)

@pytest.mark.parametrize(
        "estim_algorithm, mode, numclass, confidence_scores, class_specific", 
        list(itertools.product(moval.models.get_estim_options(),
                               ["classification", "segmentation"],
                               [3], 
                               moval.models.get_conf_options(),
                               [False, True])),
)
def test_solver(estim_algorithm, mode, numclass, confidence_scores, class_specific):

    model = moval.models.init(
            estim_algorithm,
            mode = mode,
            num_class = numclass,
            confidence_scores = confidence_scores,
            class_specific = class_specific
            )
    solver = moval.solvers.init("base-solver", model = model)
    assert isinstance(solver, moval.solvers.Solver)

    if mode == "classification":
        inp = np.random.randn(100, numclass) # 100 samples, 3 classes
        gt = np.random.randint(0, numclass, (100, ))
        _estim_acc, _estim_acc_allcls = model(inp)
        #
        solver.fit(inp, gt)
        assert model.is_fitted == True
        estim_acc, estim_acc_allcls = model(inp)

        criterions = moval.solvers.clsCalibrate(class_specific=class_specific)

        if not class_specific:
            _err = criterions(inp, gt, _estim_acc)
            err = criterions(inp, gt, estim_acc)
            if estim_algorithm != "ac-model":
                assert err <= _err
        else:
            _err = criterions(inp, gt, _estim_acc_allcls)
            err = criterions(inp, gt, estim_acc_allcls)
            if estim_algorithm != "ac-model":
                assert (err <= _err).any()
    
    if mode == "segmentation":
        # segmentation logit input
        inp = []
        gt = []
        for _ in range(5):
            inp.append(np.random.randn(numclass, 10, 10, 10)) # 5 samples, 3 classes, of volume shape ``(50, 50, 50)``
            gt.append(np.random.randint(0, numclass, (10, 10, 10)))

        _estim_acc, _estim_dsc = model(inp)
        #
        solver.fit(inp, gt)
        assert model.is_fitted == True
        estim_acc, estim_dsc = model(inp)
        
        criterions = moval.solvers.segCalibrate(class_specific=class_specific)

        if not class_specific:
            _err = criterions(inp, gt, _estim_acc)
            err = criterions(inp, gt, estim_acc)
            if estim_algorithm != "ac-model":
                assert err <= _err
        else:
            _err = criterions(inp, gt, _estim_dsc)
            err = criterions(inp, gt, estim_dsc)
            if estim_algorithm != "ac-model":
                assert (err <= _err).any()


import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
import numpy as np
from moval.solvers import register
from moval.solvers.utils import ComputMetric

class Calibrate(abc.ABC):
    """Base model for calibration criterions

    Features can be computed by calling the ``__call__()`` method. This class should not 
    be directly instaniated, and instead used as the base class for calibration criterions.

    """

    def __call__(
        self,
        inp: Union[List[Iterable], np.ndarray],
        gt: Union[List[Iterable], np.ndarray],
        estim: np.ndarray
    ) -> np.ndarray:
        """Compute the calibration error.
        """
        raise NotImplementedError()

@register("cls-calibrate")
class clsCalibrate(Calibrate):
    def __init__(self, class_specific=False):
        super().__init__()
        self.class_specific = class_specific
    
    def __call__(
        self,
        inp: np.ndarray,
        gt: np.ndarray,
        estim: np.ndarray
    ) -> np.ndarray:
        """Compute the calibration err for classification tasks.
        
        Args:
            inp: The network output (logits) of shape ``(n, d)``. 
            gt: The cooresponding annotation of shape ``(n, )``.
            estim: The estimated performance by moval.
                For class-specific version, estim is of shape ``(d, )`` which contains differences for different classes. 
        
        Returns:
            err: The difference between the estimated accuracy and the real accuracy.
                For class-specific version, err is of shape ``(d, )`` which contains differences for different classes.
        
        """

        pred = np.argmax(inp, axis = 1)

        if not self.class_specific:
            
            acc = np.sum(gt == pred) / len(gt)
            err = estim - acc

        else:

            err = np.zeros(inp.shape[1])
            for kcls in range(inp.shape[1]):
                pos_cls = np.where(pred == kcls)[0]
                acc_cls = np.sum(gt[pos_cls] == pred[pos_cls]) / len(gt[pos_cls])
                err[kcls] = estim[kcls] - acc_cls

        return np.abs(err)

@register("seg-calibrate")
class segCalibrate(Calibrate):
    def __init__(self, class_specific=False):
        super().__init__()
        self.class_specific = class_specific
    
    def __call__(
        self,
        inp: List[Iterable],
        gt: List[Iterable],
        estim: np.ndarray
    ) -> np.ndarray:
        """Compute the calibration err for segmentation tasks.
        
        Args:
            inp: The network output (logits) as a list of n ``(d, H, W, (D))`` for segmentation. 
            gt: The cooresponding annotation as a list of n ``(H, W, (D))`` for segmentation. 
            estim: The estimated performance by moval.
                For class-specific version, estim is of shape ``(d, )`` which contains differences for different classes. 
        
        Returns:
            err: The difference between the estimated DSC and the real DSC.
                For class-specific version, err is of shape ``(d, )`` which contains differences for different classes.

        """

        if not self.class_specific:
            pred_all_flatten = []
            gt_all_flatten = []
            for n_case in range(len(inp)):

                pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                gt_case     = gt[n_case] # ``(H, W, (D))``
                
                pred_all_flatten.append(pred_case.flatten())
                gt_all_flatten.append(gt_case.flatten())
            
            pred_all_flatten = np.concatenate(pred_all_flatten)
            gt_all_flatten = np.concatenate(gt_all_flatten)

            acc = np.sum(gt_all_flatten == pred_all_flatten) / len(gt_all_flatten)
            err = estim - acc

            return np.abs(err)

        else:
            
            pred_all_flatten_bg = []
            gt_all_flatten_bg = []
            dsc = []

            for n_case in range(len(inp)):

                pred_case   = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                gt_case     = gt[n_case] # ``(H, W, (D))``
                
                pos_bg = np.where(pred_case.flatten() == 0)[0]

                pred_all_flatten_bg.append(pred_case.flatten()[pos_bg])
                gt_all_flatten_bg.append(gt_case.flatten()[pos_bg])

                dsc_case = np.zeros(inp[n_case].shape[0])
                for kcls in range(1, inp[n_case].shape[0]):
                    dsc_case[kcls] = ComputMetric(pred_case == kcls, gt_case == kcls)
                dsc.append(dsc_case)

            pred_all_flatten_bg = np.concatenate(pred_all_flatten_bg)
            gt_all_flatten_bg = np.concatenate(gt_all_flatten_bg)

            acc_bg = np.sum(gt_all_flatten_bg == pred_all_flatten_bg) / len(gt_all_flatten_bg)
            
            err = estim - np.mean(np.array(dsc), axis=0)
            err[0] = estim[0] - acc_bg

            return np.abs(err)


import pytest
import numpy as np
import nibabel as nib
import itertools
from moval.solvers.utils import ComputMetric

import os
import zipfile
import gdown
import moval

"""Test the func with 3d segmentation tasks.

    Examples:
        >>> pytest tests/test_demo_seg_3d.py

"""

# download the data, which we used for MICCAI 2022

output = "data_moval.zip"
if not os.path.exists(output):
    url = "https://drive.google.com/u/0/uc?id=139pqxkG2ccIFq6qNArnFJWQ2by2Spbxt&export=download"
    output = "data_moval.zip"
    gdown.download(url, output, quiet=False)

directory_data = "data_moval"
if not os.path.exists(directory_data):
    with zipfile.ZipFile(output, 'r') as zip_ref:
        zip_ref.extractall(directory_data)

# now I am playing with prostate segmentation
Datafile_eval = "data_moval/Prostateresults/seg-eval.txt"
Imglist_eval = open(Datafile_eval)
Imglist_eval_read = Imglist_eval.read().splitlines()

logits = []
gt = []
# to accelerate the debugging speed, crop the middel 60 x 60 x 30 cub for training/inference.
for Imgname_eval in Imglist_eval_read:
    GT_file = Imgname_eval.replace("data", "data_moval")
    caseID = Imgname_eval.split("/")[-1][:6]
    logit_cls0_file = "data_moval/Prostateresults/prostateval/results/pred_" + caseID + "cls0_prob.nii.gz"
    logit_cls1_file = "data_moval/Prostateresults/prostateval/results/pred_" + caseID + "cls1_prob.nii.gz"
    logit_cls0_read = nib.load(logit_cls0_file)
    logit_cls1_read = nib.load(logit_cls1_file)
    logit_cls0      = logit_cls0_read.get_fdata()   # ``(H, W, D)``
    logit_cls1      = logit_cls1_read.get_fdata()
    GT_read         = nib.load(GT_file)
    GTimg           = GT_read.get_fdata()           # ``(H, W, D)``
    logit_cls0      = logit_cls0[logit_cls0.shape[0] //2 - 30: logit_cls0.shape[0] //2 + 30,
                                 logit_cls0.shape[1] //2 - 30: logit_cls0.shape[1] //2 + 30,
                                 logit_cls0.shape[2] //2 - 15: logit_cls0.shape[2] //2 + 15]
    logit_cls1      = logit_cls1[logit_cls1.shape[0] //2 - 30: logit_cls1.shape[0] //2 + 30,
                                 logit_cls1.shape[1] //2 - 30: logit_cls1.shape[1] //2 + 30,
                                 logit_cls1.shape[2] //2 - 15: logit_cls1.shape[2] //2 + 15]
    GTimg           = GTimg[GTimg.shape[0] //2 - 30: GTimg.shape[0] //2 + 30,
                            GTimg.shape[1] //2 - 30: GTimg.shape[1] //2 + 30,
                            GTimg.shape[2] //2 - 15: GTimg.shape[2] //2 + 15]
    logit_cls = np.stack((logit_cls0, logit_cls1))  # ``(d, H, W, D)``
    logits.append(logit_cls)
    gt.append(GTimg)

# logits is a list of length ``n``,  each element has ``(d, H, W, D)``. 
# gt is a list of length ``n``,  each element has ``(H, W, D)``.
# H, W and D could differ for different cases.

Datafile_test = "data_moval/Prostateresults/seg-testA.txt"
Imglist_test = open(Datafile_test)
Imglist_test_read = Imglist_test.read().splitlines()

logits_test = []
gt_test = []
for Imgname_test in Imglist_test_read:
    GT_file = Imgname_test.replace("data", "data_moval")

    caseID = Imgname_test.split("/")[-1][:6]

    logit_cls0_file = "data_moval/Prostateresults/prostattestcondition_A/results/pred_" + caseID + "cls0_prob.nii.gz"
    logit_cls1_file = "data_moval/Prostateresults/prostattestcondition_A/results/pred_" + caseID + "cls1_prob.nii.gz"

    logit_cls0_read = nib.load(logit_cls0_file)
    logit_cls1_read = nib.load(logit_cls1_file)
    logit_cls0      = logit_cls0_read.get_fdata()
    logit_cls1      = logit_cls1_read.get_fdata()
    GT_read         = nib.load(GT_file)
    GTimg           = GT_read.get_fdata()           # ``(H, W, D)``

    logit_cls = np.stack((logit_cls0, logit_cls1))  # ``(n, H, W, D)``
    
    logits_test.append(logit_cls)
    gt_test.append(GTimg)


# logits_test is a list of length ``n``,  each element has ``(d, H, W, D)``. 
# gt_test is a list of length ``n``,  each element has ``(H, W, D)``.
# H, W and D could differ for different cases.

results_files = "results_demo_seg_3d.txt"
# clean previous results
if os.path.isfile(results_files):
    os.remove(results_files)


# estim_algorithm = "ts-model"
# mode = "segmentation"
# confidence_scores = "max_class_probability-conf"
# class_specific = True

@pytest.mark.parametrize(
        "estim_algorithm, mode, confidence_scores, class_specific", 
        list(itertools.product(moval.models.get_estim_options(),
                               ["segmentation"],
                               moval.models.get_conf_options(),
                               [False, True])),
)
def test_seg_3d(estim_algorithm, mode, confidence_scores, class_specific):

    moval_model = moval.MOVAL(
                mode = mode,
                confidence_scores = confidence_scores,
                estim_algorithm = estim_algorithm,
                class_specific = class_specific
                )

    #
    moval_model.fit(logits, gt)
    estim_dsc = moval_model.estimate(logits)
    
    DSC_list = []
    for n_case in range(len(logits)):
        pred_case   = np.argmax(logits[n_case], axis = 0) # ``(H, W, (D))``
        gt_case     = gt[n_case] # ``(H, W, (D))``

        DSC = ComputMetric(pred_case == 1, gt_case == 1)
        DSC_list.append(DSC)
    m_DSC = np.mean(np.array(DSC_list))
    
    err_val_dsc = np.abs( m_DSC - estim_dsc )

    # save the test err in the result files.

    estim_dsc_test = moval_model.estimate(logits_test)

    DSC_list_test = []
    for n_case in range(len(logits_test)):
        pred_case   = np.argmax(logits_test[n_case], axis = 0) # ``(H, W, (D))``
        gt_case     = gt_test[n_case] # ``(H, W, (D))``

        DSC = ComputMetric(pred_case == 1, gt_case == 1)
        DSC_list_test.append(DSC)
    m_DSC_test = np.mean(np.array(DSC_list_test))

    err_test = np.abs( m_DSC_test - estim_dsc_test )

    test_condition = f"estim_algorithm = {estim_algorithm}, mode = {mode}, confidence_scores = {confidence_scores}, class_specific = {class_specific}"

    with open(results_files, 'a') as f:
        f.write(test_condition)
        f.write('\n')
        f.write("validation dsc err: ")
        f.write(str(err_val_dsc))
        f.write('\n')
        f.write("validation predicted dsc: ")
        f.write(str(estim_dsc))
        f.write('\n')
        f.write("moval parameter: ")
        f.write(str(moval_model.model_.param))
        f.write('\n')
        if moval_model.model_.extend_param:
            f.write("moval extended parameter: ")
            f.write(str(moval_model.model_.param_ext))
            f.write('\n')
        f.write("test err: ")
        f.write(str(err_test))
        f.write('\n')
        f.write('\n')


import pytest
import numpy as np

import moval
import moval.solvers

"""Test the confidence calculation functions.

    Examples:
        >>> pytest tests/test_criterions.py

"""

@pytest.mark.parametrize("class_specific", [False, True])
def test_cls_criterions(class_specific):
    criterions = moval.solvers.clsCalibrate(class_specific=class_specific)
    assert isinstance(criterions, moval.solvers.Calibrate)

    # classification logit input
    numclass = 10
    inp = np.random.randn(100, numclass) # 100 samples, 10 classes
    gt = np.random.randint(0, numclass, (100, ))
    if class_specific:
        estim = np.random.rand(numclass, )
    else:
        estim = np.random.rand(1, )[0]

    err = criterions(inp, gt, estim)
    if class_specific:
        assert len(err) == numclass
    assert np.mean(err) < 1
    assert np.mean(err) > 0

@pytest.mark.parametrize("class_specific", [False, True])
def test_seg_criterions(class_specific):
    criterions = moval.solvers.segCalibrate(class_specific=class_specific)
    assert isinstance(criterions, moval.solvers.Calibrate)

    # segmentation logit input
    numclass = 3
    inp = []
    gt = []
    for _ in range(5):
        inp.append(np.random.randn(numclass, 50, 50, 50)) # 5 samples, 3 classes, of volume shape ``(100, 100, 100)``
        gt.append(np.random.randint(0, numclass, (50, 50, 50)))
    if class_specific:
        estim = np.random.rand(numclass, )
    else:
        estim = np.random.rand(1, )[0]

    err = criterions(inp, gt, estim)
    if class_specific:
        assert len(err) == numclass
    assert np.mean(err) <= 1 and np.mean(err) >= 0



import pytest
import numpy as np

import moval

"""Test the save and load functions.

    Examples:
        >>> pytest tests/test_load.py

"""

def test_load():
    moval_model = moval.MOVAL(
                        mode = "classification",
                        confidence_scores = "max_class_probability-conf",
                        estim_algorithm = "ts-model",
                        class_specific = False
                        )
    logits = np.random.randn(1000, 10)
    gt = np.random.randint(0, 10, (1000))
    moval_model.fit(logits, gt)
    moval_model.save('./foo.pkl')
    loaded_model = moval.MOVAL.load('./foo.pkl')
    estim_acc = loaded_model.estimate(logits)
    assert estim_acc <= 1 and estim_acc >= 0


import numpy as np

def ComputMetric(ACTUAL: np.ndarray, PREDICTED: np.ndarray) -> float:
    """Caculate the DSC between prediction and the GT.

    Args:
        PREDICTED: The predicted segmentation map of shape ``((n), H, W, (D))`.`
        ACTUAL: The ground truth segmentation map of shape ``((n), H, W, (D))`.

    Returns:
        dice: A float scalar which represents the calculated dice score.
    
    """
    ACTUAL = ACTUAL.flatten()
    PREDICTED = PREDICTED.flatten()
    idxp = ACTUAL == True
    idxn = ACTUAL == False

    tp = np.sum(ACTUAL[idxp] == PREDICTED[idxp])
    tn = np.sum(ACTUAL[idxn] == PREDICTED[idxn])
    fp = np.sum(idxn) - tn
    fn = np.sum(idxp) - tp
    if tp == 0 :
        dice = 0
    else:
        dice = 2 * tp / (2 * tp + fp + fn)
    return dice


import abc
from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
import numpy as np
from typing import Any
from moval.models import register
import moval.models
from moval.models.utils import SoftDiceLoss

class Model(abc.ABC):
    """Base model for MOVAL experiments.

    Features can be computed by calling the ``__call__()`` method. This class should not 
    be directly instaniated, and instead used as the base class for MOVAl models.

    Args:
        mode (str): The given task to estimate model performance.
        num_class (int): The number of class for the given task.
        confidence_scores (str): The method to calculate confidence scores.
        class_specific (bool): If ``True``, the calculation will match class-wise confidence to class-wise accuracy/DSC.
    
    Atrributes:
        estim_algorithm (str):
            The chosen estimated algorithms for confidence calibration.
        mode (str):
            The given task to estimate model performance.
            This will choose the criterions for parameter optimization.
        num_class (int): 
            The number of class for the given task.
            This will decide the number of parameter to optimize and optimization criterions.
        confidence_scores (str): 
            The method to calculate confidence scores. 
            This will decide the methods chosen from confidences.
        class_spcific (bool):
            If ``True``, the calculation will match class-wise confidence to class-wise accuracy/DSC.
            This will also affect the number of parameter to optimize and optimization criterions.
        conf (moval.models.Confidence):
            The Confidence class to calculate the confidence scores.
        max_value (float):
            The maximum value to normalize the confidence score before calibration.
            This is only necessary when :py:attr:`conf` is unbounded.
        min_value (float):
            The minimum value to normalize the confidence score before calibration.
            This is only necessary when :py:attr:`conf` is unbounded.
        param (np.ndarray):
            The parameter to be optimized.
        extend_param (bool):
            If ``True``, the model contains more parameters (x2) and need two-stage optimization.
            Currently, this is only applicable for ``ts-atc-model``.
        is_fitted (bool):
            If ``True``, the model is fitted using validation data and ready to use for estimating performance.
        is_training (bool):
            If ``True``, the normalization paramters would be updated using the input data.

    """

    def __init__(
        self,
        estim_algorithm: str,
        mode: str,
        num_class: int,
        confidence_scores: str,
        class_specific: bool,
        extend_param: bool
    ):
        self.estim_algorithm = estim_algorithm
        self.mode = mode
        self.num_class = num_class
        self.confidence_scores = confidence_scores
        self.class_specific = class_specific
        self.is_fitted = False
        self.extend_param = extend_param

        self.conf = moval.models.init(self.confidence_scores)
        
        self.is_training = False
        
        # Note, for non-normalized method, we choose the parameter as the lower bound for normalization.
        if self.conf.normalization:
            self.max_value = 0.
            self.min_value = 0.

        if self.class_specific:
            self.param = np.ones(num_class)
        else:
            self.param = np.array([1])
        
        # allocate addtional set of parameters.
        if self.extend_param:
            if self.class_specific:
                self.param_ext = np.ones(num_class)
            else:
                self.param_ext = np.array([1])
    
    def train(self):
        """Initialize the model into training mode.
        """
        self.is_training = True
    
    def eval(self):
        """Initialize the model into evaluation mode.
        """
        self.is_training = False

    def _find_normalize(self, scores: Union[List[Iterable], np.ndarray]):
        """Find the normalization parameter for the confidence scores.

        Args:
            scores: The unbounded confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.

        Notes:
            The intensity value, max and min value, will be saved.
        
        """
        if isinstance(scores, list):
            max_value_list = []
            min_value_list = []
            for score in scores:
                max_value_list.append(np.max(score))
                min_value_list.append(np.min(score))
            self.max_value = np.max(max_value_list)
            self.min_value = np.min(min_value_list)
        
        else:
            self.max_value = np.max(scores)
            self.min_value = np.min(scores)

    def _normalize(self, scores: Union[List[Iterable], np.ndarray], lb: np.ndarray = None, pred: np.ndarray = None, e1: float = 1e-6) -> Union[List[Iterable], np.ndarray]:
        """Normalize the confidence scores to [(1-self.param), 1] if the score is not bounded.

        Note:
            This replace the temperature scaling process of this confidence scores, e.g., doctor, energy.

        Args:
            scores: The unbounded confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
            e1: A small number to prevent unexpected results of division.
            lb: Lower bound of scores to calibrate the confidence scores of shape ``(d, )`` for class-specific or ``(1, )`` for class-agnostic.
                If it is omited as None, the scores are normalized to [0, 1] (lb = 0).
            pred: The prediction derived from the logits, of shape ``(n, )`` or a list of n ``(H, W, (D))``, used for class-specific operation.
        
        Returns:
            normalized_scores: The normalized confidence scores, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        if self.is_training:
            
            self._find_normalize(scores)

        if isinstance(scores, list):
            
            _normalized_scores = []
            for n_case in range(len(scores)):
                _normalized_scores.append( (scores[n_case] - self.min_value) / (self.max_value - self.min_value + e1) )

            if lb is None:
                return _normalized_scores
            
            else:

                normalized_scores = []

                for n_case in range(len(_normalized_scores)):

                    # reshape normalized_scores, pred to vectors.
                    inp_shape = _normalized_scores[n_case].shape

                    # flaten the logit for segmentation.
                    normalized_score = _normalized_scores[n_case].reshape((-1))
                    pred_case = pred[n_case].reshape((-1))

                    # normalized_scores now is of shape ``(n, )``
                    # pred now is of shape ``(n, )``

                    if not self.class_specific:
                        normalized_score = normalized_score * self.param + (1 - self.param)
                    else:
                        # class-wise average confidence
                        for kcls in range(self.num_class):
                            pos_cls = np.where(pred_case == kcls)[0]
                            normalized_score[pos_cls] = normalized_score[pos_cls] * self.param[kcls] + (1 - self.param[kcls])
                    
                    normalized_score = normalized_score.reshape(inp_shape)

                    normalized_scores.append(normalized_score)

                return normalized_scores

        else:

            normalized_scores = (scores - self.min_value) / (self.max_value - self.min_value + e1)

            if lb is None:
                return normalized_scores
            else:
                # normalized_scores now is of shape ``(n, )``
                # pred now is of shape ``(n, )``

                if not self.class_specific:
                    normalized_scores = normalized_scores * self.param + (1 - self.param)
                else:
                    # class-wise average confidence
                    for kcls in range(self.num_class):
                        pos_cls = np.where(pred == kcls)[0]
                        normalized_scores[pos_cls] = normalized_scores[pos_cls] * self.param[kcls] + (1 - self.param[kcls])

                return normalized_scores
    
    def __call__(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False, gt_guide: np.ndarray = None) -> float:
        """Calculate the performance using network output.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.
            gt_guide: The cooresponding annotation guide of shape ``(n, d)`` for segmentation. This is only rquired for segmentation task during optimizaing.
                We will utilize this to determine if there is label in the case. If not, we do not calculate the dsc and utilize for optimizing.
                This is because we do not want to optimize parameter with blank segmentation map. 
                This should be bool, if ``False``, it means that there isn't any manuel label of class d in this sample.
        
        Returns:
            estim_acc: A float scalar that represents the estimated accuracy for the given input data.
            estim_cls: Estimated class-wise performance of shape ``(d, )``.
                For classification task, this is the class-wise estimated accuracy.
                For segmentation task, this class-wise estimated accuracy for background and DSC estimation for other foreground classes.

        """

        # calibrate the confidence scores
        if self.extend_param:
            score = self.calibrate(inp, midstage)
        else:
            score = self.calibrate(inp)

        # Estimate the performance.
        if self.mode == "classification":
            # for classification tasks, return average confidence
            estim_acc = np.mean(score)
            # class-wise average confidence
            pred = np.argmax(inp, axis = 1)
            estim_acc_allcls = []
            for kcls in range(inp.shape[1]):
                score_kcls = score[np.where(pred == kcls)[0]]
                estim_acc_kcls = np.mean(score_kcls)
                estim_acc_allcls.append(estim_acc_kcls)
            return estim_acc, np.array(estim_acc_allcls)
        elif self.mode == "segmentation":
            # for segmentation tasks, return average confidence for background class (c=0), and return the soft dsc for other foreground classes.
            scores = []
            scores_bg = []
            for n_case in range(len(inp)):
                pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                pred_flatten = pred_case.flatten() # ``n``
                score_flatten = score[n_case].flatten() # ``n``
                score_flatten_bg = score_flatten[np.where(pred_flatten == 0)[0]]
                scores.append(score_flatten)
                scores_bg.append(score_flatten_bg)

            estim_acc = np.mean( np.concatenate(scores) )
            estim_acc_bg = np.mean( np.concatenate(scores_bg) )
            
            # Note, the calculation of dice score need the probability of non-maximum classes 
            # here we extend score from shape ``(n, H, W, (D))`` to ``(n, d, H, W, (D))``, the other dimension are just filled with zeros
            estim_dsc_list = []
            for n_case in range(len(inp)):
                pred_case = np.argmax(inp[n_case], axis = 0) # ``(H, W, (D))``
                pred_flatten = pred_case.flatten() # ``n``
                score_case = score[n_case] # ``(H, W, (D))``
                score_flatten = score[n_case].flatten() # ``n``
                score_filled = np.zeros((score_flatten.shape + (self.num_class,))) # ``(n, d)``
                score_filled[np.arange(score_filled.shape[0]), pred_flatten] = score_flatten # ``(n, d)``
                score_filled = score_filled.T # ``(d, n)``
                score_filled = score_filled.reshape(((self.num_class,) + score_case.shape)) # ``(d, H, W, (D))``
                #
                estim_dsc = SoftDiceLoss(score_filled[np.newaxis, ...], pred_case[np.newaxis, ...])
                #
                if isinstance(gt_guide, np.ndarray):
                    gt_case = gt_guide[n_case]
                    # We remove the class DSC if there isn't any in gt
                    for kcls in range(len(estim_dsc)):
                        if gt_case[kcls] is False:
                            estim_dsc[kcls] = 0

                estim_dsc_list.append(estim_dsc)
            m_estim_dsc = np.mean(np.array(estim_dsc_list), axis=0)
            
            return estim_acc, np.concatenate([np.array([estim_acc_bg]), m_estim_dsc[1:]])
        else:
            raise ValueError(f"Unknown mode '{self.mode}'")

    def calibrate(self, scores: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores with parameters.

        Different estimation algorithms would adopt different strateiges to calibrate the scores.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data.
        
        """
        
        raise NotImplementedError()


@register("ac-model")
class acModel(Model):
    """MOVAL model with average confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ac-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores for average confidence.
        
        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.

        """

        # Calualte the confidence socres.
        score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            score = self._normalize(score)
        
        return score

@register("ts-model")
class tsModel(Model):
    """MOVAL model with temperature scaling."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ts-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores with temperature scaling.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        score = self.conf(inp, self.param) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            if self.mode == "classification":
                pred = np.argmax(inp, axis = 1)
                score = self._normalize(score, lb = self.param, pred = pred)
            elif self.mode == "segmentation":
                preds = []
                for n_case in range(len(score)):
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    preds.append(pred_case)
                score = self._normalize(score, lb = self.param, pred = preds)
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        
        return score

@register("doc-model")
class docModel(Model):
    """MOVAL model with difference of confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "doc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on difference of confidence.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        Note:
            The implementation of class-specific DoC for segmentation is different from what we did in our MICCAI work.
            In MICCAI, we calculate the difference as the difference of soft dsc. Here, we modify the confidence score to match dsc.
            Current version should be more compatable with other techiques.
        
        """

        # Calualte the confidence socres.
        score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            score = self._normalize(score)

        if not self.class_specific:
            if self.mode == "classification":
                return score - (1 - self.param)
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(score)):
                    score_case = score[n_case] - (1 - self.param)
                    score_post.append(score_case)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score[pos_cls] = score[pos_cls] - (1 - self.param[kcls])
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(score)):
                    score_case = score[n_case] # ``(H, W, (D))``
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    score_flatten = score_case.flatten()
                    pred_flatten = pred_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_flatten[pos_flatten_cls] = score_flatten[pos_flatten_cls] - (1 - self.param[kcls])
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(score_reshape)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")

@register("atc-model")
class atcModel(Model):
    """MOVAL model with average thresholded confidence."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "atc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = False
        )

    def calibrate(self, inp: Union[List[Iterable], np.ndarray]) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on average thresholded confidence.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        _score = self.conf(inp) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation

        # Normalize the scores, if needed.
        if self.conf.normalization:
            _score = self._normalize(_score)

        if not self.class_specific:
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                score[_score > self.param] = 1
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    score = np.zeros(_score[n_case].shape)
                    score[_score[n_case] > self.param] = 1
                    score_post.append(score)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score_pos = score[pos_cls]
                    score_pos[_score[pos_cls] > self.param[kcls]] = 1
                    score[pos_cls] = score_pos
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    _score_case = _score[n_case] # ``(H, W, (D))``
                    score_case = np.zeros(_score[n_case].shape)
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    pred_flatten = pred_case.flatten()
                    score_flatten = score_case.flatten()
                    _score_flatten = _score_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_pos = score_flatten[pos_flatten_cls]
                        score_pos[_score_flatten[pos_flatten_cls] > self.param[kcls]] = 1
                        score_flatten[pos_flatten_cls] = score_pos
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(score_reshape)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")

@register("ts-atc-model")
class tsatcModel(Model):
    """MOVAL model with average thresholded confidence after temperature scaling."""

    def __init__(self, mode: str, num_class: int, confidence_scores: str, class_specific: bool):
        super().__init__(
            estim_algorithm = "ts-atc-model",
            mode = mode,
            num_class = num_class,
            confidence_scores = confidence_scores,
            class_specific = class_specific,
            extend_param = True
        )
    
    def calibrate(self, inp: Union[List[Iterable], np.ndarray], midstage: bool = False) -> Union[List[Iterable], np.ndarray]:
        """Calibrate the confidence scores based on average thresholded confidence after temperature scaling.

        Args:
            inp: The network output (logits) of shape ``(n, d)`` or a list of n ``(d, H, W, (D))``.
            midstage: If ``True``, return the first calibrated results.

        Returns:
            calibrated_scores: The calibrated scores which would match the accuracy/DSC on validation data, of shape ``(n, )`` or a list of n ``(H, W, (D))``.
        
        """

        # Calualte the confidence socres.
        _score = self.conf(inp, self.param) # ``(n,)`` for classification or a list of n ``(H, W, (D))`` for segmentation
        
        # Normalize the scores, if needed.
        if self.conf.normalization:
            if self.mode == "classification":
                pred = np.argmax(inp, axis = 1)
                _score = self._normalize(_score, lb = self.param, pred = pred)
            elif self.mode == "segmentation":
                preds = []
                for n_case in range(len(_score)):
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    preds.append(pred_case)
                _score = self._normalize(_score, lb = self.param, pred = preds)
            else:
                raise ValueError(f"Unknown mode '{self.mode}'") 

        if midstage:
            return _score

        if not self.class_specific:
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                score[_score > self.param_ext] = 1
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    score = np.zeros(_score[n_case].shape)
                    score[_score[n_case] > self.param_ext] = 1
                    score_post.append(score)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")
        else:
            # class-wise average confidence
            if self.mode == "classification":
                score = np.zeros(inp.shape[0])
                pred = np.argmax(inp, axis = 1)
                for kcls in range(inp.shape[1]):
                    pos_cls = np.where(pred == kcls)[0]
                    score_pos = score[pos_cls]
                    score_pos[_score[pos_cls] > self.param_ext[kcls]] = 1
                    score[pos_cls] = score_pos
                return score
            elif self.mode == "segmentation":
                score_post = []
                for n_case in range(len(_score)):
                    _score_case = _score[n_case] # ``(H, W, (D))``
                    score_case = np.zeros(_score[n_case].shape)
                    pred_case = np.argmax(inp[n_case], axis = 0)
                    score_shape = score_case.shape
                    pred_flatten = pred_case.flatten()
                    score_flatten = score_case.flatten()
                    _score_flatten = _score_case.flatten()
                    #
                    for kcls in range(self.num_class):
                        pos_flatten_cls = np.where(pred_flatten == kcls)[0]
                        score_pos = score_flatten[pos_flatten_cls]
                        score_pos[_score_flatten[pos_flatten_cls] > self.param_ext[kcls]] = 1
                        score_flatten[pos_flatten_cls] = score_pos
                        score_reshape = score_flatten.reshape(score_shape) # ``(H, W, (D))``
                    score_post.append(score_reshape)
                return score_post
            else:
                raise ValueError(f"Unknown mode '{self.mode}'")

        return score


import pytest
import numpy as np
import itertools

import moval
import moval.models
import moval.registry

"""Test the confidence calculation functions.

    Examples:
        >>> pytest tests/test_models.py

"""

def test_registry():
    assert moval.registry.is_registry(moval.models)
    assert moval.registry.is_registry(moval.models, check_docs=True)

@pytest.mark.parametrize(
        "estim_algorithm, mode, numclass, confidence_scores, class_specific", 
        list(itertools.product(moval.models.get_estim_options(),
                               ["classification", "segmentation"],
                               [2, 10],
                               moval.models.get_conf_options(),
                               [False, True])),
)
def test_model(estim_algorithm, mode, numclass, confidence_scores, class_specific):
    model = moval.models.init(
            estim_algorithm,
            mode = mode,
            num_class = numclass,
            confidence_scores = confidence_scores,
            class_specific = class_specific
            )
    assert isinstance(model, moval.models.Model)

    model.train()
    if mode == "classification":
        # classification logit input
        inp = np.random.randn(100, numclass) # 100 samples
        estim_acc, estim_acc_allcls = model(inp)
        assert estim_acc <= 1 and estim_acc >= 0
        assert len(estim_acc_allcls) == numclass
        assert np.max(estim_acc) <= 1 and np.min(estim_acc) >= 0
    
    if mode == "segmentation":
        # segmentation logit input
        inp = []
        for _ in range(5):
            inp.append(np.random.randn(numclass, 50, 50, 50)) # 10 samples, 3 classes, of volume shape ``(100, 100, 100)``
        estim_acc, estim_dsc = model(inp)
        assert estim_acc <= 1 and estim_acc >= 0
        assert len(estim_dsc) == numclass
        assert np.max(estim_dsc) <= 1 and np.min(estim_dsc) >= 0